"""
Theming support for Django's collectstatic functionality.
See https://docs.djangoproject.com/en/1.8/ref/contrib/staticfiles/
"""
from __future__ import absolute_import

import os.path

from django.conf import settings
from django.contrib.staticfiles.storage import StaticFilesStorage
from django.utils._os import safe_join

import theming
from theming.thread_locals import get_current_theme


class ThemeStorage(StaticFilesStorage):
    """
    Comprehensive theme aware Static files storage.
    """
    # prefix for file path, this prefix is added at the beginning of file path before saving static files during
    # collectstatic command.
    # e.g. having "test-theme" as prefix will cause files to be saved as "test-theme/images/logo.png"
    # instead of "images/logo.png"
    prefix = None

    def __init__(self, location=None, base_url=None, file_permissions_mode=None,
                 directory_permissions_mode=None, prefix=None):

        self.prefix = prefix

        super(ThemeStorage, self).__init__(
            location=location,
            base_url=base_url,
            file_permissions_mode=file_permissions_mode,
            directory_permissions_mode=directory_permissions_mode,
        )

    def url(self, name):
        """
        Returns url of the asset, themed url will be returned if the asset is themed otherwise default
        asset url will be returned.

        Arguments:
            name: name of the asset, e.g. 'images/logo.png'

        Returns:
            url of the asset, e.g. '/static/red-theme/images/logo.png' if current theme is red-theme and logo
            is provided by red-theme otherwise '/static/images/logo.png'
        """
        prefix = ''
        theme = get_current_theme()

        # get theme prefix from site address if if asset is accessed via a url
        if theme:
            prefix = theme.name

        # get theme prefix from storage class, if asset is accessed during collectstatic run
        elif self.prefix:
            prefix = self.prefix

        # join theme prefix with asset name if theme is applied and themed asset exists
        if prefix and self.themed(name, prefix):
            name = os.path.join(prefix, name)

        return super(ThemeStorage, self).url(name)

    def themed(self, name, theme_name):
        """
        Returns True if given asset override is provided by the given theme otherwise returns False.

        Arguments:
            name: asset name e.g. 'images/logo.png'
            theme_name: theme name e.g. 'red-theme', 'test-theme'

        Returns:
            True if given asset override is provided by the given theme otherwise returns False
        """
        if not theming.is_enabled():
            return False

        # in debug mode check static asset from within the project directory
        if settings.DEBUG:
            themes_location = theming.get_base_dir(theme_name)
            # Nothing can be themed if we don't have a theme location or required params.
            if not all((themes_location, theme_name, name)):
                return False

            themed_path = "/".join([
                themes_location,
                theme_name,
                "static/"
            ])
            name = name[1:] if name.startswith("/") else name
            path = safe_join(themed_path, name)
            return os.path.exists(path)

        # in live mode check static asset in the static files dir defined by "STATIC_ROOT" setting
        return self.exists(os.path.join(theme_name, name))



"""
Module to hold all the nuts and bolts for theming a django site.
"""
from __future__ import absolute_import

from .core import get_base_dir, get_themes, is_enabled

__version__ = "0.1.0"


default_app_config = "theming.apps.ThemingConfig"  # pylint: disable=invalid-name



"""
Wrapper for loading templates from "templates" directories for themes.
"""
from __future__ import absolute_import

from django.template.loaders.filesystem import Loader as FilesystemLoader

from theming import get_themes
from theming.thread_locals import get_current_request, get_current_theme


class Loader(FilesystemLoader):
    """
    Theming aware template loader.
    """
    def get_dirs(self):
        """
        Return all template directories including theme template directories.

        Returns:
            (list): A list containing template directories.
        """
        # TODO: Cache result for faster performance
        dirs = super(Loader, self).get_dirs()
        theme_dirs = []

        if get_current_request():
            # If the template is being loaded in a request, prepend the current theme's template directories
            # so the theme's templates take precedence.
            theme = get_current_theme()

            if theme:
                theme_dirs = theme.template_dirs
        else:
            # If we are outside of a request, we should load all directories for all themes.
            for theme in get_themes():
                theme_dirs.extend(
                    theme.template_dirs
                )

        return theme_dirs + dirs



"""
Module containing admin interfaces for the theming app.
"""
from __future__ import absolute_import

from django.contrib import admin

from .models import Theme


@admin.register(Theme)
class ThemeAdmin(admin.ModelAdmin):
    """
    Admin UI definitions for `Theme` model.
    """

    list_display = ('site', 'name')
    search_fields = ('site__domain', 'name')



"""
Module to contain all of the exceptions defined and used by the theming app.
"""
from __future__ import absolute_import

from django.core.exceptions import ImproperlyConfigured as DjangoImproperlyConfigured


class ThemeException(Exception):
    """
    Base exception for all theming related exceptions.
    """


class MiddlewareNotActivated(ThemeException):
    """
    Exception raised when features being used depends on a theming middleware.
    """


class ImproperlyConfigured(ThemeException, DjangoImproperlyConfigured):
    """
    Theming is somehow improperly configured.
    """



# -*- coding: utf-8 -*-
"""
Thread scoped variables.
"""
from __future__ import absolute_import

from threading import local

from theming.exceptions import MiddlewareNotActivated

__thread_locals__ = local()


def set_thread_variable(key, value):
    """
    Set thread variable with the given key, value pair.

    Arguments:
        key (str): Name of the thread variable.
        value (object): Value of the thread variable.

    Example:
        >>> set_thread_variable('current_theme_name', "test-theme")
    """
    setattr(__thread_locals__, key, value)


def get_thread_variable(key, default=None):
    """
    Get thread variable with the given key. Return `default` if key not found in thread data.

    Arguments:
        key (str): Name of the thread variable.
        default (object): Value to return if variable with given key is not present in thread data.

    Example:
        >>> get_thread_variable('current_theme_name', "default-theme")
        "default-theme"
        >>> set_thread_variable('current_theme_name', "test-theme")
        >>> get_thread_variable('current_theme_name', "default-theme")
        "test-theme"
    """
    return getattr(__thread_locals__, key, default)


def set_request_variable(key, value):
    """
    Set request variable with the given key, value pair.

    Arguments:
        key (str): Name of the request variable.
        value (object): Value of the request variable.

    Example:
        >>> set_request_variable('current_theme_name', "test-theme")
    """
    request = get_current_request()
    if not request:
        raise MiddlewareNotActivated(
            "Unable to get request object. Make sure CurrentRequestMiddleware is installed",
        )

    return setattr(request, key, value)


def get_request_variable(key, default=None):
    """
    Get request variable with the given key. Return `default` if key not found in thread data.

    Arguments:
        key (str): Name of the request variable.
        default (object): Value to return if variable with given key is not present in request data.

    Example:
        >>> get_request_variable('current_theme_name', "default-theme")
        "default-theme"
        >>> set_request_variable('current_theme_name', "test-theme")
        >>> get_request_variable('current_theme_name', "default-theme")
        "test-theme"
    """
    request = get_current_request()
    if not request:
        raise MiddlewareNotActivated(
            "Unable to get request object. Make sure CurrentRequestMiddleware is installed",
        )

    return getattr(request, key, default)


def set_current_request(request):
    """
    Save request object in the thread data.

    Arguments:
        request (HttpRequest): Django request object.
    """
    return set_thread_variable('request', request)


def get_current_request(default=None):
    """
    Get request object from the thread data associated with the current django request.

    Arguments:
        default (object): Default value to return in case there is no current request being processed.
    """
    return get_thread_variable('request', default)


def set_current_theme(theme):
    """
    Save theme object in the request data.

    Arguments:
        theme (theming.models.Theme): Theme object
    """
    return set_request_variable('theme', theme)


def get_current_theme(default=None):
    """
    Get theme object from the thread data associated with the current django request.

    Arguments:
        default (object):  Default value to return in case there is no current request being processed.
    """
    return get_request_variable('theme', default)



"""
App Config for theming.
"""
from __future__ import absolute_import

import os

from django.apps import AppConfig
from django.conf import settings

from theming.exceptions import ImproperlyConfigured


class ThemingConfig(AppConfig):
    """
    Basic config class for theming app.
    """

    name = 'theming'

    def ready(self):
        """
        Perform validations on theme settings.

        settings must contains the `THEMING` setting with the following format.
        ```
            THEMING = {
                'ENABLED': True,  # boolean indicating whether theming feature is enabled of disabled.
                'DIRS': [],  # List of absolute paths (str) to the directories that contain themes.
            }
        ```
        """
        if not hasattr(settings, 'THEMING'):
            raise ImproperlyConfigured(
                '"THEMING" setting not set in django settings file. '
                'If you are not using theming then remove it from INSTALLED_APPS.'
            )
        if 'DIRS' not in settings.THEMING:
            raise ImproperlyConfigured(
                'The THEMING["DIRS"] setting must be populated.',
            )
        if 'ENABLED' not in settings.THEMING:
            raise ImproperlyConfigured(
                'The THEMING["ENABLED"] setting must be populated.',
            )
        if not isinstance(settings.THEMING['DIRS'], (list, tuple)):
            raise ImproperlyConfigured(
                'The THEMING["DIRS"] setting is not a tuple or list. Perhaps you forgot a trailing comma?',
            )
        if not all([isinstance(theme_dir, str) for theme_dir in settings.THEMING['DIRS']]):
            raise ImproperlyConfigured(
                'THEMING["DIRS"] must contain only string paths.',
            )
        if not all([theme_dir.startswith("/") for theme_dir in settings.THEMING['DIRS']]):
            raise ImproperlyConfigured(
                'THEMING["DIRS"] must contain only absolute paths to themes dirs.',
            )
        if not all([os.path.isdir(theme_dir) for theme_dir in settings.THEMING['DIRS']]):
            raise ImproperlyConfigured(
                'THEMING["DIRS"] must contain valid paths.',
            )



"""
Module to contain middleware defined by the theming app.
"""
from __future__ import absolute_import

from django.utils.deprecation import MiddlewareMixin

from theming.models import Theme
from theming.thread_locals import set_current_request, set_current_theme


class CurrentRequestMiddleware(MiddlewareMixin):
    """
    Middleware that sets `request` attribute in the local threading storage.
    """

    def process_request(self, request):  # pylint: disable=no-self-use
        """
        Save request object in local thread data.
        """
        set_current_request(request)


class CurrentThemeMiddleware(MiddlewareMixin):
    """
    Middleware that sets `theme` attribute to request object.

    Dependencies:
        This middleware is dependant on the following middleware,
        middleware must only be added after these two in the `MIDDLEWARE` setting.
            1. (django.contrib.sites.middleware.CurrentSiteMiddleware)
            2. (theming.middleware.CurrentRequestMiddleware)
    """

    def process_request(self, request):  # pylint: disable=no-self-use
        """
        Save theme object in local thread data.
        """
        set_current_theme(Theme.get_theme(request.site))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-06 08:35
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='Theme',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('site', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='theme', to='sites.Site')),
            ],
        ),
    ]



"""
Static file finders for Django.
https://docs.djangoproject.com/en/1.8/ref/settings/#std:setting-STATICFILES_FINDERS
Yes, this interface is private and undocumented, but we need to access it anyway.
"""
from __future__ import absolute_import

import os
from collections import OrderedDict

from django.conf import settings
from django.contrib.staticfiles import utils
from django.contrib.staticfiles.finders import BaseFinder
from django.core.checks import Error
from django.utils import six

import theming
from theming.static.storage import ThemeStorage


class ThemeFilesFinder(BaseFinder):
    """
    A static files finder that looks in the directory of each theme as
    specified in the source_dir attribute.
    """
    storage_class = ThemeStorage
    source_dir = 'static'

    def __init__(self, *args, **kwargs):
        # The list of themes that are handled
        self.themes = []
        # Mapping of theme names to storage instances
        self.storages = OrderedDict()

        if theming.is_enabled():
            themes = theming.get_themes()
            for theme in themes:
                theme_storage = self.storage_class(
                    os.path.join(theme.path, self.source_dir),
                    prefix=theme.name,
                )

                self.storages[theme.name] = theme_storage
                if theme.name not in self.themes:
                    self.themes.append(theme.name)

        super(ThemeFilesFinder, self).__init__(*args, **kwargs)

    def check(self, **kwargs):  # pylint: disable=unused-argument
        """
        Check if current instance data has errors.
        """
        errors = []
        if 'DIRS' not in settings.THEMING:
            errors.append(Error(
                'The THEMING["DIRS"] setting must be populated.',
                id='theming.static.E001',
            ))
        elif not isinstance(settings.THEMING['DIRS'], (list, tuple)):
            errors.append(Error(
                'The THEMING["DIRS"] setting is not a tuple or list.',
                hint='Perhaps you forgot a trailing comma?',
                id='theming.static.E002',
            ))
        elif not all([isinstance(theme_dir, str) for theme_dir in settings.THEMING['DIRS']]):
            errors.append(Error(
                'THEMING["DIRS"] must contain only strings.',
                id='theming.static.E003',
            ))
        elif not all([theme_dir.startswith("/") for theme_dir in settings.THEMING['DIRS']]):
            errors.append(Error(
                'THEMING["DIRS"] must contain only absolute paths to themes dirs.',
                id='theming.static.E004',
            ))
        elif not all([os.path.isdir(theme_dir) for theme_dir in settings.THEMING['DIRS']]):
            errors.append(Error(
                'THEMING["DIRS"] must contain valid paths.',
                id='theming.static.E005',
            ))
        return errors

    def list(self, ignore_patterns):
        """
        List all files in all theme storages.
        """
        for storage in six.itervalues(self.storages):
            if storage.exists(''):  # check if storage location exists
                for path in utils.get_files(storage, ignore_patterns):
                    yield path, storage

    def find(self, path, all=False):  # pylint: disable=redefined-builtin
        """
        Looks for files in the theme directories.
        """
        matches = []
        theme_dir = path.split("/", 1)[0]

        themes = {t.name: t for t in theming.get_themes()}
        # if path is prefixed by theme name then search in the corresponding storage other wise search all storages.
        if theme_dir in themes:
            theme = themes[theme_dir]
            path = "/".join(path.split("/")[1:])
            match = self.find_in_theme(theme.name, path)
            if match:
                if not all:
                    return match
                matches.append(match)
        return matches

    def find_in_theme(self, theme_name, path):
        """
        Find a requested static file in an theme's static locations.
        """
        storage = self.storages.get(theme_name, None)
        if storage:
            # only try to find a file if the source dir actually exists
            if storage.exists(path):
                matched_path = storage.path(path)
                if matched_path:
                    return matched_path
        return None



"""
Core logic for Theming.
"""
from __future__ import absolute_import, division

import os

from path import Path

from django.conf import settings


def is_enabled():
    """
    Return `True` if theming is enabled, return `False` otherwise.
    """
    return settings.THEMING.get('ENABLED', False)


def is_theme_dir(_dir):
    """
    Return true if given dir is a theme directory, returns False otherwise.

    A theme dir must have subdirectory 'static' or 'templates' or both.

    Arguments:
        _dir: directory path to check for a theme

    Returns:
        Returns true if given dir is a theme directory.
    """
    theme_sub_directories = {'static', 'templates'}
    return bool(os.path.isdir(_dir) and theme_sub_directories.intersection(os.listdir(_dir)))


def get_theme_base_dirs():
    """
    Return a list of all directories that contain themes.

    Example:
        >> get_theme_base_dirs()
        ['/var/themes/']

    Returns:
         (list): list of theme base directories
    """
    return [Path(theme_dir) for theme_dir in settings.THEMING['DIRS']]


def get_base_dir(name):
    """
    Return absolute path to the directory that contains the given theme.

    Arguments:
        name (str): theme directory name to get base path for

    Returns:
        (str): Base directory that contains the given theme
    """
    # TODO: handle theme not found
    for themes_dir in get_theme_base_dirs():
        if name in (_dir for _dir in os.listdir(themes_dir) if is_theme_dir(themes_dir / _dir)):
            return themes_dir

    # TODO: raise ThemeNotFound instead of returning ''
    return ''


def get_all_theme_template_dirs():
    """
    Return a list of all template directories, for all the themes.

    Example:
        >> get_all_theme_template_dirs()
        [
            '/var/themes//test-theme/templates/',
        ]

    Returns:
        (list): list of directories containing theme templates.
    """
    template_paths = list()

    for theme in get_themes():
        template_paths.extend(
            theme.template_dirs
        )
    return template_paths


def get_themes(themes_dir=None):
    """
    Return a list of all themes known to the system.

    If themes_dir is given then return only the themes residing inside that directory.

    Arguments:
        themes_dir (str): (Optional) Path to themes base directory

    Returns:
        list of themes known to the system.
    """
    # TODO: Fix this
    # Importing here to avoid circular import
    from theming.models import Theme

    themes_dirs = [Path(themes_dir)] if themes_dir else get_theme_base_dirs()
    # pick only directories and discard files in themes directory
    themes = []

    for _dir in themes_dirs:
        themes.extend(
            [Theme(name=name) for name in os.listdir(_dir) if is_theme_dir(Path(_dir) / name)]
        )

    return themes



"""
Module to contain model definitions for theming app.
"""
from __future__ import absolute_import

from path import Path

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

import theming


class Theme(models.Model):
    """
    Django ORM model for Theme db table.

    Fields:
        site (ForeignKey): Foreign Key field pointing to django Site model
        theme_dir_name (CharField): Contains directory name for any site's theme (e.g. 'red-theme')
    """

    site = models.OneToOneField(Site, related_name='theme', on_delete=models.CASCADE)
    name = models.CharField(max_length=255)

    def __eq__(self, other):
        """
        Return True if given theme is same as the self.

        Arguments:
            other: Theme object to compare with self

        Returns:
            (bool) True if two themes are the same else False
        """
        return (self.name, self.path) == (other.name, other.path)

    def __hash__(self):
        """
        Return unique hash for the theme.
        """
        return hash((self.name, self.path))

    def __str__(self):
        """
        Return unique readable string for the theme.
        """
        return u"<Theme: {name} at '{path}'>".format(name=self.name, path=self.path)

    def __repr__(self):
        """
        Return unique readable string for the theme.
        """
        return self.__str__()

    @property
    def base_dir(self):
        """
        Return base directory path for the theme.
        """
        # TODO: handle theme not found
        return Path(theming.get_base_dir(str(self.name)))

    @property
    def path(self):
        """
        Return full directory path for the theme.
        """
        return self.base_dir / self.name

    @property
    def template_dirs(self):
        """
        Return list of all template directories of the theme.
        """
        return [
            self.path / 'templates',
        ]

    @staticmethod
    def get_theme(site):
        """
        Return SiteTheme object for given site.

        Return default site theme if theme for the given site can not be found
        and `DEFAULT_SITE_THEME` setting has a proper value.

        Arguments:
            site (django.contrib.sites.models.Site): site object related to the current site.

        Returns:
            SiteTheme object for given site or a default site set by `DEFAULT_SITE_THEME`
        """
        try:
            return site.theme
        except ObjectDoesNotExist:
            if settings.THEMING.get('DEFAULT', None):
                return Theme(site=site, name=settings.THEMING['DEFAULT'])
            return None
